/*
 *
 *  hcisdpfunc - Bluetooth functions for Linux
 *
 *  Copyright (C) 2006-2008  Sean Godsell <sgodsell@hotmail.com>
 *
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#include <stdio.h>
#include <errno.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <sys/socket.h>

#include <bluetooth/bluetooth.h>
#include <bluetooth/hci.h>
#include <bluetooth/hci_lib.h>
#include <bluetooth/sdp.h>
#include <bluetooth/sdp_lib.h>

#include <netinet/in.h>



/*
 * Convert a string to a BDADDR, with a few "enhancements"
 */
char *fixbaddr(char *str)
{
    char tmp[255] = {0};
    char *p = str;
    int  o=0,i=0;
    while(*str != '\0') {
        if (i>0 && i%2 ==  0 && *str != ':') {
            tmp[o] = ':';
            o++;
        }
        tmp[o] = *str;
        i++;
        o++;
        str++;
    }
    strcpy(p, tmp);
    return p;
}

/*
 * Convert a string to a BDADDR, with a few "enhancements"
 */
int estr2ba(char *str, bdaddr_t *ba)
{
    /* Only trap "local", "any" is already dealt with */
    if(!strcmp(str, "local")) {
        bacpy(ba, BDADDR_LOCAL);
        return 0;
    }
    return str2ba(str, ba);
}

/* Pass args to the inquiry/search handler */
struct search_context {
	char		*svc;		/* Service */
	uuid_t		group;		/* Browse group */
	int		view;		/* View mode */
	uint32_t	handle;		/* Service record handle */
};

typedef int (*handler_t)(bdaddr_t *bdaddr, struct search_context *arg);

static char UUID_str[MAX_LEN_UUID_STR];
static bdaddr_t interface;

/* Definition of attribute members */
struct member_def {
	char *name;
};

/* Definition of an attribute */
struct attrib_def {
	int			num;		/* Numeric ID - 16 bits */
	char			*name;		/* User readable name */
	struct member_def	*members;	/* Definition of attribute args */
	int			member_max;	/* Max of attribute arg definitions */
};

/* Definition of a service or protocol */
struct uuid_def {
	int			num;		/* Numeric ID - 16 bits */
	char			*name;		/* User readable name */
	struct attrib_def	*attribs;	/* Specific attribute definitions */
	int			attrib_max;	/* Max of attribute definitions */
};

/* Context information about current attribute */
struct attrib_context {
	struct uuid_def		*service;	/* Service UUID, if known */
	struct attrib_def	*attrib;	/* Description of the attribute */
	int			member_index;	/* Index of current attribute member */
};


/* Allow us to do nice formatting of the lists */
//static char *indent_spaces = "                                         ";

int foundDN = 0;
int foundCN = 0;
int saveChannel = -1;




/* Same for the UUIDs. See BT assigned numbers */
static struct uuid_def uuid16_names[] = {
	/* -- Protocols -- */
	{ 0x0001, "SDP", NULL, 0 },
	{ 0x0003, "RFCOMM", NULL, 0 },
	{ 0x1103, "DialupNetworking (DUN)", NULL, 0 },
	{ 0x1204, "GenericTelephony", NULL, 0 },
};

const int uuid16_max = sizeof(uuid16_names)/sizeof(struct uuid_def);



void get_service_class(void *value, void *userData)
{
	char ServiceClassUUID_str[MAX_LEN_SERVICECLASS_UUID_STR];
	uuid_t *uuid = (uuid_t *)value;

	sdp_uuid2strn(uuid, UUID_str, MAX_LEN_UUID_STR);
	sdp_svclass_uuid2strn(uuid, ServiceClassUUID_str, MAX_LEN_SERVICECLASS_UUID_STR);
	// Look specifically for Dialup networking (1103)
	if (uuid->type != SDP_UUID128 && strcmp("1103", UUID_str) == 0) {
		foundDN = 1;
        }
}

void get_service_desc(void *value, void *user)
{
	char str[MAX_LEN_PROTOCOL_UUID_STR];
	sdp_data_t *p = (sdp_data_t *)value;
	int i = 0, proto = 0;
	uuid_t *uuid = (uuid_t *)value;
	sdp_uuid2strn(uuid, UUID_str, MAX_LEN_UUID_STR);

	for (; p; p = p->next, i++) {
		switch (p->dtd) {
		case SDP_UUID16:
		case SDP_UUID32:
		case SDP_UUID128:
			sdp_uuid2strn(&p->val.uuid, UUID_str, MAX_LEN_UUID_STR);
			sdp_proto_uuid2strn(&p->val.uuid, str, sizeof(str));
			proto = sdp_uuid_to_proto(&p->val.uuid);
			break;
		case SDP_UINT8:
			if (proto == RFCOMM_UUID && foundDN > 0) {
				saveChannel = p->val.uint8;
			}
			break;
		}
	}
}


void get_access_protos(void *value, void *userData)
{
	sdp_list_t *protDescSeq = (sdp_list_t *)value;
	sdp_list_foreach(protDescSeq, get_service_desc, 0);
}


/*
 * Parse a SDP record in user friendly form.
 */
void get_service_attr(sdp_record_t *rec)
{
	sdp_list_t *list = 0, *proto = 0;


	if (sdp_get_service_classes(rec, &list) == 0) {
		sdp_list_foreach(list, get_service_class, 0);
		sdp_list_free(list, free);
	}
	if (sdp_get_access_protos(rec, &proto) == 0) {
		sdp_list_foreach(proto, get_access_protos, 0);
		sdp_list_foreach(proto, (sdp_list_func_t)sdp_list_free, 0);
		sdp_list_free(proto, 0);
		foundDN = 0;
	}
	if (sdp_get_lang_attr(rec, &list) == 0) {
		sdp_list_free(list, free);
	}
	if (sdp_get_profile_descs(rec, &list) == 0) {
		sdp_list_free(list, free);
	}
}



/*
 * Show how to get a specific SDP record by its handle.
 * Not really useful to the user, just show how it can be done...
 */
int get_service(bdaddr_t *bdaddr, struct search_context *context, int quite)
{
	sdp_list_t *attrid;
	uint32_t range = 0x0000ffff;
	sdp_record_t *rec;
	sdp_session_t *session = sdp_connect(&interface, bdaddr, SDP_RETRY_IF_BUSY);

	if (!session) {
		char str[20];
		ba2str(bdaddr, str);
		return -1;
	}

	attrid = sdp_list_append(0, &range);
	rec = sdp_service_attr_req(session, context->handle, SDP_ATTR_REQ_RANGE, attrid);
	sdp_list_free(attrid, 0);
	sdp_close(session);

	if (!rec) {
		if (!quite) {
			return -1;
		} else
			return 0;
	}

	/* Display user friendly form */
	get_service_attr(rec);

	sdp_record_free(rec);
	return 0;
}



/*
 * Request possible SDP service records
 */
int get_Channel(char *addr)
{
	struct search_context context;
	uint32_t base[] = { 0x10000, 0x1002e, 0x110b, 0x90000, 0x2008000, 0x4000000, 0x100000, 0x1000000 };
	bdaddr_t bdaddr;
	int i, n, err = 0, num = 16;

	/* Initialise context */
	memset(&context, '\0', sizeof(struct search_context));

	if (addr == NULL)
		return -1;

	/* Convert command line parameters */
	estr2ba(addr, &bdaddr);

	for (i = 0; i < sizeof(base) / sizeof(uint32_t); i++)
		for (n = 0; n < num; n++) {
			context.handle = base[i] + n;
			err = get_service(&bdaddr, &context, 1);
			if (err < 0 || saveChannel > -1)
				goto done;
		}

done:
	return saveChannel;
}


int get_hciscan_list(int display, char *findstr, char *baddr, char *dname)
{

    inquiry_info *ii = NULL;
    int max_rsp, num_rsp;
    int dev_id, sock, len, flags;
    int i,x;
    char addr[254] = { 0 };
    char name[254] = { 0 };
    inquiry_info  *devclass;
    int  dclass;
    int  found = 0;

    dev_id = hci_get_route(NULL);
    sock = hci_open_dev( dev_id );
    if (dev_id < 0 || sock < 0) {
        printf("Error opening bluetooth device!\n");
	return 0;
    }

    len  = 8;
    max_rsp = 255;
    flags = IREQ_CACHE_FLUSH;
    ii = (inquiry_info*)malloc(max_rsp * sizeof(inquiry_info));
    memset(ii, 0, max_rsp * sizeof(inquiry_info));

    num_rsp = hci_inquiry(dev_id, len, max_rsp, NULL, &ii, flags);
    if( num_rsp < 0 )
        perror("hci_inquiry");


    for (i = 0; i < num_rsp; i++) {
        ba2str(&(ii+i)->bdaddr, addr);
        devclass = (ii+i);
        // Get the device class and mask off the last byte
        dclass = *((int *) devclass->dev_class) & 0xffffff;
        // Look for a phone class
//        if (dclass & 0x200)
        {
            memset(name, 0, sizeof(name));
            if (hci_read_remote_name(sock, &(ii+i)->bdaddr, sizeof(name),
                name, 0) < 0)
                strcpy(name, "[unknown]");

            if (!display) {
                /* check if the argument matches a device name
                 */
                if (strncasecmp(findstr, name, strlen(findstr)) == 0) {
                    strcpy(baddr, addr);
                    strcpy(dname, name);
                    found = 1;
                }
                /* Check if argument 1 is an address
                 */
                if (strncasecmp(findstr, addr, strlen(findstr)) == 0) {
                    strcpy(baddr, addr);
                    strcpy(dname, name);
                    found = 1;
                }
            }
            if (display)
                printf("%s  %s\n", addr, name);

        }
    }
    if (display)
        printf("\n");

    free( ii );
    close( sock );
    return found;

}


#ifdef USE_MAIN_SDP

int main(int argc, char *argv[])
{
	int i, opt, channel;

	bacpy(&interface, BDADDR_ANY);

	channel = get_Channel(argv[1]);

	printf("Channel = %d\n", channel);

	return 1;
}

#endif

